VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "TextFile"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False

' TextFile:
' VBA class module for creating UTF-8 textfiles
'
' Usage:
' Dim fd As New TextFile
' fd.FileCreate "c:\any\path\to\the\textfile", "UTF-8"
' fd.TextWrite "any text"
' fd.TextWriteLine "any text line"
' fd.FileClose
'
' You can also specify "UTF-16" for UTF-16 LE encoding, and other for your local encoding.
'
' Todo:
' Functions for reading such textfiles.
' (but I don't need this feature now, so ... :-P)
'
' Author: Hiroto Kagotani <kagotani@cne.okayama-u.ac.jp>
' Date: 2004-03-03
' Copyright: This is free software with absolutely no warranty.
'            You are permitted to use/copy/modify/redistribute this software freely.

Private propEncoding As String
Private propIsOpen As Boolean
Private propTextStream As Object
Private propFileNumber As Integer

Public Function FileCreate(path As String, enc As String) As Boolean
    Dim fs As Object
    
    On Error GoTo HandleError
    
    propEncoding = enc
    
    If Me.IsOpen Then
        Me.FileClose
    End If
    
    If Me.Encoding = "UTF-8" Then
        propFileNumber = FreeFile
        
        ' truncate
        Open path For Output As propFileNumber
        Close propFileNumber
        
        Open path For Binary Access Write As propFileNumber
    Else
        Set fs = CreateObject("Scripting.FileSystemObject")
        If Me.Encoding = "UTF-16" Then
            Set propTextStream = fs.CreateTextFile(path, True, True)
        Else
            Set propTextStream = fs.CreateTextFile(path, True)
        End If
    End If

    propIsOpen = True
    FileCreate = True
ExitProc:
    Exit Function

HandleError:
    FileCreate = False
    Resume ExitProc
End Function

Public Function FileClose() As Boolean
    On Error GoTo HandleError
        
    If Me.IsOpen Then
        If Me.Encoding = "UTF-8" Then
            Close propFileNumber
        Else
            propTextStream.Close
        End If
    End If
    
    propIsOpen = False
    FileClose = True
ExitProc:
    Exit Function

HandleError:
    FileClose = False
    Resume ExitProc
End Function

Public Function TextWrite(s As String) As Boolean
    On Error GoTo HandleError
    
    If Me.Encoding = "UTF-8" Then
        Put propFileNumber, , String2Utf8(s)
    Else
        If Me.Encoding = "UTF-16" Then
        Else
            s = StrConv(StrConv(s, vbFromUnicode), vbUnicode)
        End If
        propTextStream.Write s
    End If
    
    TextWrite = True
ExitProc:
    Exit Function

HandleError:
    TextWrite = False
    Resume ExitProc
End Function

Public Function TextWriteLine(s As String) As Boolean
    TextWriteLine = Me.TextWrite(s & vbCrLf)
End Function

Property Get IsOpen() As Boolean
    IsOpen = propIsOpen
End Property

Property Get Encoding() As String
    Encoding = propEncoding
End Property

#Const UCS4 = False

Public Function String2Utf8(s As String) As Byte()
    Dim b() As Byte
    Dim blength As Integer
    Dim bindex As Integer
    Dim ch As Long
    
    ' ftp://ftp.rfc-editor.org/in-notes/rfc2279.txt
    
    ' U+0000 - U+007F                             00000000.0zzzzzzz: 0zzzzzzz
    ' U+0080 - U+07FF                             00000yyy.yyzzzzzz: 110yyyyy 10zzzzzz
    ' U+0800 - U+FFFF                             xxxxyyyy.yyzzzzzz: 1110xxxx 10yyyyyy 10zzzzzz
    ' U+0001 0000 - U+001F FFFF 00000000.000wwwxx.xxxxyyyy.yyzzzzzz: 11110www 10xxxxxx 10yyyyyy 10zzzzzz
    ' U+0020 0000 - U+03FF FFFF 000000vv.wwwwwwxx.xxxxyyyy.yyzzzzzz: 111110vv 10wwwwww 10xxxxxx 10yyyyyy 10zzzzzz
    ' U+0400 0000 - U+7FFF FFFF 0uvvvvvv.wwwwwwxx.xxxxyyyy.yyzzzzzz: 1111110u 10vvvvvv 10wwwwww 10xxxxxx 10yyyyyy 10zzzzzz
    
    blength = 0
    For sindex = 1 To Len(s)
        ch = AscW(Mid(s, sindex, 1))
        If ch < 0 Then
            ch = ch + 65536
        End If
        If ch < &H80 Then
            blength = blength + 1
        ElseIf ch < &H800 Then
            blength = blength + 2
        ElseIf ch < &H10000 Then
            blength = blength + 3
#If UCS4 = True Then
        ElseIf ch < &H200000 Then
            blength = blength + 4
        ElseIf ch < &H4000000 Then
            blength = blength + 5
        ElseIf ch < &H80000000 Then
            blength = blength + 6
#End If
        Else
            ' unsupported
        End If
    Next sindex
    ReDim b(0 To blength - 1) As Byte
    bindex = 0
    For sindex = 1 To Len(s)
        ch = AscW(Mid(s, sindex, 1))
        If ch < 0 Then
            ch = ch + 65536
        End If
        If ch < &H80 Then
            b(bindex) = ch And &H7F
            bindex = bindex + 1
        ElseIf ch < &H800 Then
            b(bindex) = &HC0 Or ((ch And &H7C0) \ &H40)
            b(bindex + 1) = &H80 Or ch And &H3F
            bindex = bindex + 2
        ElseIf ch < &H10000 Then
            b(bindex) = &HE0 Or ((ch And &HF000) \ &H1000)
            b(bindex + 1) = &H80 Or ((ch And &HFC0) \ &H40)
            b(bindex + 2) = &H80 Or ch And &H3F
            bindex = bindex + 3
#If UCS4 = True Then
        ElseIf ch < &H200000 Then
            b(bindex) = &HF0 Or ((ch And &H1C0000) \ &H40000)
            b(bindex + 1) = &H80 Or ((ch And &H3F000) \ &H1000)
            b(bindex + 2) = &H80 Or ((ch And &HFC0) \ &H40)
            b(bindex + 3) = &H80 Or ch And &H3F
            bindex = bindex + 4
        ElseIf ch < &H4000000 Then
            b(bindex) = &HF8 Or ((ch And &H3000000) \ &H1000000)
            b(bindex + 1) = &H80 Or ((ch And &HFC0000) \ &H40000)
            b(bindex + 2) = &H80 Or ((ch And &H3F000) \ &H1000)
            b(bindex + 3) = &H80 Or ((ch And &HFC0) \ &H40)
            b(bindex + 4) = &H80 Or ch And &H3F
            bindex = bindex + 5
        ElseIf ch < &H80000000 Then
            b(bindex) = &HFC Or ((ch And &H40000000) \ &H40000000)
            b(bindex + 1) = &H80 Or ((ch And &H3F000000) \ &H1000000)
            b(bindex + 2) = &H80 Or ((ch And &HFC0000) \ &H40000)
            b(bindex + 3) = &H80 Or ((ch And &H3F000) \ &H1000)
            b(bindex + 4) = &H80 Or ((ch And &HFC0) \ &H40)
            b(bindex + 5) = &H80 Or ch And &H3F
            bindex = bindex + 6
#End If
        Else
            ' unsupported
        End If
    Next sindex

    String2Utf8 = b
    
End Function
